package com.ef_prime.rflow;

import java.awt.Component;
import java.io.File;

import javax.swing.JOptionPane;

import com.ef_prime.rflow.flow.FlowSetting;
import com.ef_prime.rflow.flow.FlowState;
import com.ef_prime.rflow.flow.GraphWriter;
import com.ef_prime.rflow.flow.ViewLayout;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.io.ElementWriter;
import com.ef_prime.rflow.io.FlowDataWriter;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.tool.OptionDialog;
import com.ef_prime.rflow.util.ExtensionUtility;
import com.ef_prime.rflow.util.FileControl;
import com.ef_prime.rflow.util.RandomStringUtility;

/**
 * 
 * フローフォルダ関連の処理メソッド群
 *
 */
public class FlowFolderSupport {
	
	/** フローフォルダのランダム名の文字数 **/
	public static final int RANDOM_LENGTH = 10;
	
	/**
	 * 追加データがあるかを返します
	 * @param state
	 * @return
	 */
	public static boolean isAdditionalData(FlowState state){
		return state.getCacheNodes().size() > 0;
	}
	
	/**
	 * フロー設定にフローフォルダ設定されていない場合にランダム名に設定します。
	 * @param flowSetting
	 * @param workingFile 
	 */
	public static void setFolderDefault(FlowSetting flowSetting, File workingFile){
		if(workingFile == null)return;
		String folder = (String)flowSetting.getProperty(FlowSetting.Key.FOLDER);
		if(folder.equals((String)FlowSetting.Key.FOLDER.getDefaultValue())){
			setNewFolder(flowSetting, workingFile, false);
		}
	}
	
	/**
	 * フロー設定に新しいファイル用にフローフォルダを設定します。
	 * 基本的にランダム名でフォルダを設定します。すでに、使用されている場合は、もう一度ランダム名を生成します。
	 * @param flowSetting フォルダを設定するフロー設定
	 * @param file 新しいファイル
	 * @param acceptIfNameEquals 名前が重複してもネームファイルが一致するならそのフォルダを使う
	 */
	public static void setNewFolder(FlowSetting flowSetting, File file, boolean acceptIfNameEquals){
		String newFolder = RandomStringUtility.getRandomAlphanumeric(RANDOM_LENGTH);
		File fNewFolder = new File(file.getParent()+"/"+Constants.RFLOW_FOLDER+"/"+newFolder); //$NON-NLS-1$ //$NON-NLS-2$
		while(fNewFolder.exists()){
			if(acceptIfNameEquals){
				String s = Constants.getFileName(fNewFolder);
				if(s != null && s.equals(ExtensionUtility.removeExtension(file)))
					break;							
			}
			newFolder = RandomStringUtility.getRandomAlphanumeric(RANDOM_LENGTH);
			fNewFolder = new File(file.getParent()+"/"+Constants.RFLOW_FOLDER+"/"+newFolder); //$NON-NLS-1$ //$NON-NLS-2$
		}
		flowSetting.setProperty(FlowSetting.Key.FOLDER, fNewFolder.getName());
	}
	
	/**
	 * ネームファイルとファイル名の一致を調べます。
	 * 一致しない場合、処理の選択ダイアログを表示し、選択された動作を行います。
	 * @param rflow
	 * @param layout 読み込まれたViewLayout。保存用に使用されます。
	 * @return false if process canceled
	 */
	public static boolean checkFlowFolder(Rflow rflow, ViewLayout layout, FlowState state){

		File workFile = rflow.getWorkingFile();
		File orginalFlowDir = rflow.getFlowDir();
		FlowSetting flowSetting = rflow.getFlowSetting();
		
		
		String nameFileName = Constants.getFileName(orginalFlowDir);
		if(nameFileName == null)return true;
		String fileName = ExtensionUtility.removeExtension(workFile);
		if(!nameFileName.equals(fileName)){
			Integer selected = OptionDialog.showOptionDialog(rflow.frame, Messages.getString("Message.confirm"),  //$NON-NLS-1$
					Messages.getString("Message.unlinked_flow_cache"),  //$NON-NLS-1$
					new String[]{Messages.getString("Message.unlinked_option_copy"),Messages.getString("Message.unlinked_option_alter"),Messages.getString("Message.unlinked_option_do_nothing")}); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			if(selected == null)//canceled
				return false;
			switch(selected){
			case 0://コピー
				//重複名チェック
				setNewFolder(flowSetting, workFile, false);
				File newFlowDir = rflow.getFlowDir();
				if(!newFlowDir.mkdirs())break;
				for(File f: orginalFlowDir.listFiles())
					rflow.fileControl.copyFile(f, newFlowDir, FileControl.OVERWRITE_NO, false);
				//ネームファイルを変更
				if(!renameNameFile(rflow.frame, flowSetting, nameFileName, fileName, newFlowDir))
					return false;
				//保存
				save(workFile, layout, flowSetting, rflow.frame);
				break;
			case 1://変更
				//ネームファイルを変更
				if(!renameNameFile(rflow.frame, flowSetting, nameFileName, fileName, orginalFlowDir))
					return false;
				break;
			case 2://使用しない （新規作成）
			default:
				setNewFolder(flowSetting, workFile, false);
				//すべてのキャッシュをクリア
				for(FlowNode node : state.getCacheNodes())
					node.getCacheData().setValid(false);
				//保存
				save(workFile, layout, flowSetting, rflow.frame);
				break;
			}
		}
		return true;
	}

	/**
	 * ネームファイルをリネームします。
	 * 失敗した場合、falseを返します。
	 * @param parent
	 * @param flowSetting
	 * @param oldName
	 * @param newName
	 * @param dir
	 * @return false if failed
	 */
	private static boolean renameNameFile(Component parent, FlowSetting flowSetting, String oldName, String newName, File dir) {
		File nf = new File(dir, Constants.NAME_FILE_PREFIX+oldName);
		if(!nf.renameTo(new File(dir, Constants.NAME_FILE_PREFIX+newName)))
		{
			JOptionPane.showMessageDialog(parent, Messages.getString("Error.file.rename")+": "+nf.getPath(), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			return false;
		}
		return true;
	}
	
	private static void save(File file, ViewLayout layout, FlowSetting setting, Component parent){
		ElementWriter[] elementWriters = new ElementWriter[]{
				new GraphWriter(layout, setting)
				};
		try {
			new FlowDataWriter().write(file, elementWriters);
		} catch (Exception e) {
			if(Rflow.DEBUG > 0)e.printStackTrace();
			JOptionPane.showMessageDialog(parent, Messages.getString("Error.save"), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
		} 
	}
	
	/**
	 * 与えられたフローフォルダが空の場合消去します。
	 * @param flowFolder
	 */
	public static void cleanFlowDir(FileControl fileControl, File flowFolder){
		if(flowFolder.exists()){
			File cacheFolder = new File(flowFolder, Constants.CACHE_DIR);
			File tempFolder = new File(flowFolder, Constants.FILE_TEMP_DIR);
			boolean noCache = !cacheFolder.exists() || cacheFolder.listFiles().length == 0;
			boolean noTemp = !tempFolder.exists() || tempFolder.listFiles().length == 0;
			if(noCache && noTemp){
				fileControl.delete_file(flowFolder, false);
			}
		}
	}
}
