package com.ef_prime.rflow.flow;

import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.Set;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import com.ef_prime.rflow.RCommand;
import com.ef_prime.rflow.Rflow;
import com.ef_prime.rflow.Setting;
import com.ef_prime.rflow.console.AbstractCommand;
import com.ef_prime.rflow.console.Command;
import com.ef_prime.rflow.console.CommandAdapter;
import com.ef_prime.rflow.console.CommandReceiver;
import com.ef_prime.rflow.console.CompoundCommand;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.base.SubflowNodeModel;
import com.ef_prime.rflow.node.base.TunnelNodeModel;
import com.ef_prime.rflow.resource.Messages;

public class FlowExecuter {

	private Rflow rflow;
	private FlowViewer viewer;
	private CommandReceiver receiver;
	
	public FlowExecuter(Rflow rflow, FlowViewer viewer, CommandReceiver receiver){
		this.rflow = rflow;
		this.viewer = viewer;
		this.receiver = receiver;
	}
	
	public static boolean isSelectionExectuable(Set<FlowNode> selected){
		return getExecutableSelectionInfo(selected) != null;
	}
	
	public static NodeSelectionInfo getExecutableSelectionInfo(Set<FlowNode> selected){
		if(selected.isEmpty())return null;
		if(FlowUtil.extractCluster(selected).size() > 1)return null;
		FlowNode src = FlowUtil.getSource(selected.iterator().next(), selected);
		Set<FlowNode> ends = FlowUtil.getEndOfNodes(src, selected);
		if(ends.size() > 1)return null;
		FlowNode end = ends.iterator().next();
		
		return new NodeSelectionInfo(null, end, src);
	}

	public boolean execute(final FlowNode node, final FlowNode startNode, boolean execRest, boolean clearObjects, boolean useCache){
		if (node == null)
			return false;
		
		Command c = getCommand(node, startNode, execRest, clearObjects, useCache);
		if(c == null)return false;
		c.addCommandListener(new CommandAdapter(){
			@Override
			public void commandFinished(Command c) {
				if(startNode == null){
					viewer.getRootFlowState().setFlagNode(
							(node.getModel() == TunnelNodeModel.outModel)?node.getGraph().getParentNode():node);
				}
				rflow.explorer.getFileExplorer().update();
				rflow.explorer.getObjectBrowser().update();
				rflow.backupThread.doBackup();
			}
			@Override
			public void commandCanceled(Command c) {
				rflow.explorer.getObjectBrowser().update();
			}
			
			@Override
			public void errorOccurred(Command c) {
				rflow.explorer.getObjectBrowser().update();
			}
		});
		RCommand.getReceiver().passCommand(c);

		return true;
	}
	
	public Command getCommand(
			FlowNode invokeNode,
			FlowNode startNode,
			boolean execRest,  
			boolean clearObjects,
			boolean useCache) 
	{
		CompoundCommand command = new CompoundCommand();

		//exec parent, if out node
		if(invokeNode.getModel() == TunnelNodeModel.outModel)
			invokeNode = invokeNode.getGraph().getParentNode();
		
		if(!execRest && clearObjects){
			command.add(RCommand.getClearObjectsCommand());
		}
	
		List<FlowNode> stream = FlowUtil.getStream(invokeNode, true, false);
		
		int startIndex = 0;
		boolean cacheHit = false;
	
		if(execRest || useCache){
			FlowState state = viewer.getRootFlowState();
			//get next index of flag node or cache node
			for(int i = stream.size()-1; i >= 0; i--){
				FlowNode node = stream.get(i);
				//check flag first
				if(execRest && state.isFlagNode(node)){
					startIndex = i+1;
					break;			
				}
				
				if(useCache && node.isCacheData() && node.getCacheData().isValid()){
					startIndex = i+1;
					cacheHit = true;
					break;
				}
					
			}
		}
		
		if(useCache && cacheHit){
			final FlowNode cacheNode = stream.get(startIndex-1);
			try {
				command.add(viewer.getCacheHandler().loadCache(cacheNode));
			} catch (FileNotFoundException e) {
				JOptionPane.showMessageDialog(viewer, Messages.getString("Message.node_name")+" : "+cacheNode.getTitle()+"\n"+ e.getMessage()+"\n\n"+Messages.getString("Error.run_again"), //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
						Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
				return null;
			}
		}
		
		//start node
		if(startNode != null){
			if(startNode.getModel().isOpenSubflow()){
				startNode = ((SubflowNodeModel)startNode.getModel()).getIn();
			}
			int index = stream.indexOf(startNode);
			if(index > startIndex){
				startIndex = index;
			}
		}
		
		boolean checkError = (Boolean)Rflow.setting.getProperty(Setting.Key.CHECK_ERROR);
		for(int i = startIndex; i < stream.size(); i++){
			final FlowNode node = stream.get(i);
			NodeModel model = node.getModel();
			if(model.hasCommand() && !model.isOpenSubflow()){
				Command c = model.getCommandObject();
				c.addCommandListener(new CommandAdapter(){
					@Override
					public void commandStart(Command c) {
						setExecutingNode(node, true);
					}
				});
				if(checkError)
					c.setCheckError(true);
				command.add(c);
			}
			if(useCache && clearObjects && node.isCacheData() && !node.getCacheData().isValid()){
				try {
					command.add((AbstractCommand)viewer.getCacheHandler().cacheNode(node));
				} catch (IOException e) {
					JOptionPane.showMessageDialog(viewer, Messages.getString("Error.file.create_folder")+" : "+viewer.getCacheHandler().getCacheDir(true),//$NON-NLS-1$ //$NON-NLS-2$
							Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
					return null;
				}
			}
		}
		
		command.addCommandListener(new CommandAdapter(){
			@Override
			public void commandCanceled(Command c) {
				setExecutingNode(null, false);
				rflow.console.executionFinished();
			}
			@Override
			public void commandFinished(Command c) {
				setExecutingNode(null, true);
				rflow.console.executionFinished();
			}
		});
		
		return command;
	}
	
	private void setExecutingNode(final FlowNode node, boolean wait){
		try {
			if(wait){
				SwingUtilities.invokeAndWait(new Runnable(){
					public void run() {
						viewer.getRootFlowState().setExecutingNode(node);
						viewer.repaintFamiliy();
					}
				});
			}else{
				SwingUtilities.invokeLater(new Runnable(){
					public void run() {
						viewer.getRootFlowState().setExecutingNode(node);
						viewer.repaintFamiliy();
					}
				});
			}
		} catch (InterruptedException e) {
			if(Rflow.DEBUG > 0)e.printStackTrace();
		} catch (InvocationTargetException e) {
			if(Rflow.DEBUG > 0)e.printStackTrace();
		}
	}

	public void excuteInLocal(final FlowNode invokeNode){		
		List<FlowNode> stream = FlowUtil.getStream(invokeNode);
		FlowState state = viewer.getRootFlowState();

		//フラグノードの次のノードインデックスを取得
		int startIndex = 0;
		for(FlowNode node : stream){
			startIndex++;
			if(state.isFlagNode(node))
				break;			
		}
		if(startIndex >= stream.size())startIndex = 0;
		
		for(int i = startIndex; i < stream.size(); i++){
			FlowNode node = stream.get(i);
			String command = node.getModel().getCommand();
			receiver.passCommand(RCommand.getLocalExcutionCode(command));
			//RCommand.getR().eval(RCommand.getLocalExcutionCode(command));
			//RCommand.getR().jriWriteConsole(command);
		}
	}
	
	public void copyCommandToClipBoard(FlowNode invokeNode){
		String command = FlowUtil.getFlowCommand(invokeNode);
		Clipboard clip = Toolkit.getDefaultToolkit().getSystemClipboard();
		StringSelection s = new StringSelection(command);
		clip.setContents(s, s);
	}
}
