package com.ef_prime.rflow.flow;

import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.swing.JOptionPane;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import com.ef_prime.rflow.Constants;
import com.ef_prime.rflow.exception.ModelCreationFailedException;
import com.ef_prime.rflow.flow.model.FlowEdge;
import com.ef_prime.rflow.flow.model.FlowGraph;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.group.DefaultAppearanceGroup;
import com.ef_prime.rflow.io.ElementReader;
import com.ef_prime.rflow.io.SingleElementReader;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeModelController;
import com.ef_prime.rflow.node.NodeModelManager;
import com.ef_prime.rflow.node.NodeProperty;
import com.ef_prime.rflow.node.base.FreeNodeModel;
import com.ef_prime.rflow.resource.Messages;

import edu.uci.ics.jung.graph.Vertex;

public class GraphReader extends ElementReader implements GraphElements{
	
	private FlowGraph graph = new FlowGraph();
	private Map<Vertex, Point2D> locationMap = new HashMap<Vertex, Point2D>();
	private Dimension size;
	private boolean loc_type_absolute = false;
	
	private Point offset = new Point(0,0);
	
	private Map<Integer, Vertex> idToNodeMap = new HashMap<Integer, Vertex>();
	private Map<String, String> nodeInfoMap = new HashMap<String, String>();
	private FlowSetting flowSettings = new FlowSetting();
	private FlowState state = new FlowState();
	private Properties properties = new Properties();
	private List<ViewLayout> subflows = new ArrayList<ViewLayout>();
	private boolean inOption = false;
	private NodeModel currentModel;
	
	private GraphReader parentReader;
	private boolean isOverlay;
	private GraphReader overlayReader = null;
	
	
	public GraphReader(final Frame frame, Dimension viewerSize){
		this(frame, viewerSize, null);
	}		
		
	private GraphReader(final Frame frame, Dimension viewerSize, GraphReader parentReader){
		size = viewerSize;//default for old relative location type
		this.parentReader = parentReader;

		addOverlaySingleElementReader(GRAPH, new GraphElementReader(frame));
		addOverlaySingleElementReader(NODE, new NodeReader());
		addOverlaySingleElementReader(COMMAND, new CommandReader());
		addOverlaySingleElementReader(PROPERTY, new PropertyReader());
		addOverlaySingleElementReader(COMMENT, new CommentReader());
		addOverlaySingleElementReader(CACHE, new CacheReader());
		addOverlaySingleElementReader(ENTRY, new EntryReader());
		addOverlaySingleElementReader(OPTION, new OptionReader());
		addOverlaySingleElementReader(SUBFLOW, new SubflowReader(frame));
		addOverlaySingleElementReader(EDGE, new EdgeReader());
	}
	
	public ViewLayout getViewLayout(){
		if(!loc_type_absolute){
			for(Map.Entry<Vertex, Point2D> entry : locationMap.entrySet()){
				Point2D p = entry.getValue();
				Point2D p2 = new Point2D.Double(size.getWidth()*p.getX(), size.getHeight()*p.getY());
				locationMap.put(entry.getKey(), p2);
			}
		}
		return new ViewLayout(graph, locationMap, size, offset);
	}

	public FlowSetting getFlowSettings() {
		if (flowSettings.isEmpty())
			return null;
		return flowSettings;
	}
	
	public FlowState getFlowState(){
		return state;
	}
	
	void addOverlaySingleElementReader(String element, SingleElementReader elementReader) {
		addSingleElementReader(element, new OverlaySingleElementReader(element, elementReader));
	}
	
	private final class GraphElementReader extends SingleElementReader {
		private final Frame parent;

		private GraphElementReader(Frame parent) {
			this.parent = parent;
		}

		@Override
		public void startElement(Attributes attrs) throws SAXException {
			double current = Double.valueOf(GRAPH_FILE_VERSION);
			String version = attrs.getValue(VERSION);
			if(version != null){
				double file = Double.valueOf(attrs.getValue(VERSION));
				if (current < file){//ファイルがプログラムより新しい場合
					int ret = JOptionPane.showConfirmDialog(parent, Messages.getString("Error.new_type_file"), Messages.getString("Error.title.warning"), JOptionPane.OK_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
					if(ret != JOptionPane.OK_OPTION)
						throw new SAXException();
				}
			}
			
			String width = attrs.getValue(WIDTH);
			String height = attrs.getValue(HEIGHT);
			if(width != null && height != null)
				size = new Dimension(Integer.valueOf(width), Integer.valueOf(height));
			String loc_type = attrs.getValue(LOCATION_TYPE);
			loc_type_absolute = (loc_type != null && loc_type.equals(LOCATION_TYPE_ABSOLUTE));
			
			String x = attrs.getValue(LOCATION_OFFSET_X);
			String y = attrs.getValue(LOCATION_OFFSET_Y);
			
			if(x != null && y != null)
				offset = new Point(Integer.valueOf(x), Integer.valueOf(y));
		}
		
		@Override
		public void endElement(String text) throws SAXException {			
			if(parentReader != null){
				parentReader.isOverlay = false;
			}
		}
	}

	private final class NodeReader extends SingleElementReader {
		@Override
		public void startElement(Attributes attrs) {
			nodeInfoMap.clear();
			nodeInfoMap.put(ID, attrs.getValue(ID));
			nodeInfoMap.put(TITLE, attrs.getValue(TITLE));//ない可能性もあり。そのときnullで問題ない。
			nodeInfoMap.put(X, attrs.getValue(X));
			nodeInfoMap.put(Y, attrs.getValue(Y));
			nodeInfoMap.put(CACHE, attrs.getValue(CACHE));
		}

		@Override
		public void endElement(String text) {
			//プロパティ
			NodeProperty property = NodeProperty.createDefaultProperty();
			property.setTitle(nodeInfoMap.get(TITLE));
			String shape = nodeInfoMap.get(SHAPE);
			if(shape != null){
				DefaultAppearanceGroup dap = DefaultAppearanceGroup.valueOf(shape);
				if(dap != null)
					property.setAppearanceGroup(dap.getAppearanceGroup());
			}
			property.setComment(nodeInfoMap.get(COMMENT));
			
			
			//node
			FlowNode node = new FlowNode(currentModel, property);
			graph.addVertex(node);
			
			//id
			idToNodeMap.put(Integer.valueOf(nodeInfoMap.get(ID)), node);
		
			//位置
			if(!loc_type_absolute){
				//相対位置
				double x = Double.valueOf(nodeInfoMap.get(X));
				double y = Double.valueOf(nodeInfoMap.get(Y));
				locationMap.put(node, new Point2D.Double(x, y));
			}else{
				//絶対位置
				int x = Integer.valueOf(nodeInfoMap.get(X));
				int y = Integer.valueOf(nodeInfoMap.get(Y));
				locationMap.put(node, new Point2D.Float(x, y));
			}

			//cache
			String file = nodeInfoMap.get(FILE);
			if(file != null){
				//String hash = nodeInfoMap.get(HASH);
				String date = nodeInfoMap.get(DATE);
				String sSize = nodeInfoMap.get(SIZE);
				String gray = nodeInfoMap.get(GRAY);
				boolean isValid = (gray == null) || !Boolean.parseBoolean(gray);
				long size = (sSize == null)?0L:Long.parseLong(sSize);
				File cache = new File(file + Constants.RDATA_EXTENSION);
				state.setCache(node, new CacheData(cache, null, CacheData.parseDate(date), size, isValid));
			}
		}
	}

	private final class CommandReader extends SingleElementReader {
		@Override
		public void endElement(String text) {
			nodeInfoMap.put(COMMAND, text);
		}
	}

	private final class PropertyReader extends SingleElementReader {
		@Override
		public void startElement(Attributes attrs) {
			String title = attrs.getValue(TITLE);
			/*
			 * このnullチェックは、nodeタグにtitleがある古い仕様
			 * に対応するためのものです。
			 */
			if(title != null)
				nodeInfoMap.put(TITLE, title);
			nodeInfoMap.put(SHAPE, attrs.getValue(SHAPE));
		}
	}

	private final class CommentReader extends SingleElementReader {
		@Override
		public void endElement(String text) {
			nodeInfoMap.put(COMMENT, text);
		}
	}

	private final class CacheReader extends SingleElementReader {
		@Override
		public void startElement(Attributes attrs) throws SAXException {
			nodeInfoMap.put(FILE, attrs.getValue(FILE));
			nodeInfoMap.put(HASH, attrs.getValue(HASH));
			nodeInfoMap.put(DATE, attrs.getValue(DATE));
			nodeInfoMap.put(SIZE, attrs.getValue(SIZE));
			nodeInfoMap.put(GRAY, attrs.getValue(GRAY));
		}
	}

	private final class EntryReader extends SingleElementReader {
		private String key;

		@Override
		public void startElement(Attributes attrs) {
			key = attrs.getValue(KEY);
		}

		@Override
		public void endElement(String text) {
			if(inOption)
				properties.setProperty(key, text);
			else
				flowSettings.setProperty(key, text);
		}
	}

	private final class OptionReader extends SingleElementReader {

		@Override
		public void startElement(Attributes attrs) {
			String type = attrs.getValue(TYPE);
			nodeInfoMap.put(TYPE, type);
			properties.clear();
			subflows.clear();
			inOption = true;
		}

		@Override
		public void endElement(String text) {
		
			inOption = false;
			
			NodeModelController controller = NodeModelManager.getModelController(nodeInfoMap.get(TYPE));
			if(controller == null)
				controller = NodeModelManager.getModelController(FreeNodeModel.class);
			try {
				ViewLayout[] subflowArray = subflows.toArray(new ViewLayout[0]);
				currentModel = controller.createModel(properties, 
						nodeInfoMap.get(COMMAND), subflowArray);
				//currentModel = controller.createModel(d, nodeInfoMap.get(TITLE), nodeInfoMap.get(COMMAND));
			} catch (ModelCreationFailedException e) {
				currentModel = new FreeNodeModel(nodeInfoMap.get(COMMAND));
			}
		}
	}
	

	private final class SubflowReader extends SingleElementReader {
		
		private final Frame parent;

		private SubflowReader(Frame parent) {
			this.parent = parent;
		}

		@Override
		public void startElement(Attributes attrs) {
			isOverlay = true;
			overlayReader = new GraphReader(parent, null, GraphReader.this);
		}

		@Override
		public void endElement(String text) {
			subflows.add(overlayReader.getViewLayout());
			state.add(overlayReader.getFlowState());
		}
	}

	private final class EdgeReader extends SingleElementReader {
		@Override
		public void startElement(Attributes attrs) {
			Vertex from = idToNodeMap.get(Integer.valueOf(attrs.getValue(FROM)));
			Vertex to = idToNodeMap.get(Integer.valueOf(attrs.getValue(TO)));
			if(from == null || to == null)return;
			FlowEdge edge = new FlowEdge(from, to); 
			graph.addEdge(edge);
		}
	}

	private class OverlaySingleElementReader extends SingleElementReader{
		
		private String qName;
		private SingleElementReader reader;
		
		
		
		public OverlaySingleElementReader(String qName, SingleElementReader reader) {
			super();
			this.qName = qName;
			this.reader = reader;
		}

		@Override
		public void startElement(Attributes attrs) throws SAXException {
			if(!isOverlay)
				reader.startElement(attrs);
			else{
				overlayReader.getSingleElementReader(qName).startElement(attrs);
			}
		}

		@Override
		public void endElement(String text) throws SAXException {
			if(!isOverlay)
				reader.endElement(text);
			else{
				overlayReader.getSingleElementReader(qName).endElement(text);
			}
		}
	}
}
