package com.ef_prime.rflow.flow;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import com.ef_prime.rflow.flow.model.FlowEdge;
import com.ef_prime.rflow.flow.model.FlowGraph;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.group.DefaultAppearanceGroup;
import com.ef_prime.rflow.io.ElementWriter;
import com.ef_prime.rflow.io.FlowDataWriter;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeModelController;
import com.ef_prime.rflow.node.NodeModelManager;
import com.ef_prime.rflow.node.NodeProperty;
import com.ef_prime.rflow.node.base.SubflowNodeModel;
import com.ef_prime.rflow.util.ExtensionUtility;
import com.ef_prime.rflow.xml.XMLException;
import com.ef_prime.rflow.xml.XMLWriter;

import edu.uci.ics.jung.graph.Vertex;

public class GraphWriter implements ElementWriter, GraphElements{

	private ViewLayout layout;
	private FlowGraph graph;
	private FlowSetting flowSettings;
	
	public GraphWriter(ViewLayout layout, FlowSetting flowSettings){
		this.layout = layout;
		this.graph = this.layout.getGraph();
		this.flowSettings = flowSettings;
		if(this.flowSettings != null){
			this.flowSettings.update();
		}
	}
	
	public void write(XMLWriter out) throws XMLException {
		if(graph == null)
			return;
		
		//node -> id MAP エッジ保存用
		Map<Vertex, Integer> nodeToIdMap = new HashMap<Vertex, Integer>();
		
		Dimension size = getSize();
		
		out.writeStartElement(GRAPH);
		out.writeAttribute(VERSION, GRAPH_FILE_VERSION);
		if(size != null){
			out.writeAttribute(WIDTH, String.valueOf(size.width));
			out.writeAttribute(HEIGHT, String.valueOf(size.height));
		}
		out.writeAttribute(LOCATION_TYPE, getLocationType());
		Point offset = getLayoutOffset();
		if(offset != null){
			out.writeAttribute(LOCATION_OFFSET_X, String.valueOf(offset.x));
			out.writeAttribute(LOCATION_OFFSET_Y, String.valueOf(offset.y));
		}
		//save layouttrans
		
		//ファイルごと設定
		writeFlowSettings(out);
		
		for(Object ob_v : graph.getVertices()){
			FlowNode node = (FlowNode)ob_v;
			NodeModel nodeModel = node.getModel();
			
			int id = nodeToIdMap.size();
			nodeToIdMap.put(node, id);
			
			Point2D p = getLocation(node);
			
			out.writeStartElement(NODE);
			out.writeAttribute(ID, String.valueOf(id));
			
			//if(nodeModel.isTitle())
			//	out.writeAttribute(TITLE, nodeModel.getTitle());
			
			/*相対位置
			out.writeAttribute(X, String.valueOf((float)p.getX()));
			out.writeAttribute(Y, String.valueOf((float)p.getY()));
			*/
			out.writeAttribute(X, String.valueOf((int)p.getX()));
			out.writeAttribute(Y, String.valueOf((int)p.getY()));
			
			
			out.writeStartElement(COMMAND);
			out.writeCharacters(nodeModel.getCommand());
			out.writeEndElement();
			
			//プロパティ
			NodeProperty property = node.getProperty();
			out.writeStartElement(PROPERTY);
			if(property.getTitle() != null){
				out.writeAttribute(TITLE, property.getTitle());
			}
			if(property.getAppearanceGroup() != null){
				out.writeAttribute(SHAPE, DefaultAppearanceGroup.valueOf(property.getAppearanceGroup()).toString());
			}
			if(property.getComment() != null){
				out.writeStartElement(COMMENT);
				out.writeCharacters(property.getComment());
				out.writeEndElement();//comment
			}
			//cache
			if(node.isCacheData()){
				CacheData cacheData = node.getCacheData();
				out.writeStartElement(CACHE);
				out.writeAttribute(FILE, ExtensionUtility.removeExtension(cacheData.getFile().getName()));
				//if(cacheData.isHash())out.writeAttribute(HASH, cacheData.getHash());
				if(cacheData.isDate())out.writeAttribute(DATE, CacheData.toString(cacheData.getDate()));
				if(cacheData.isSize())out.writeAttribute(SIZE, Long.toString(cacheData.getSize()));
				if(!cacheData.isValid())out.writeAttribute(GRAY, Boolean.toString(true));
				out.writeEndElement();
			}
			out.writeEndElement();//property
			
	
			out.writeStartElement(OPTION);
			out.writeAttribute(TYPE, nodeModel.getClass().getCanonicalName());
			
			//node option
			NodeModelController controller = NodeModelManager.getModelController(nodeModel.getClass());
			Properties properties = null;
			if(controller != null)
				properties = controller.getOptionProperties(nodeModel);					
			
			if(properties != null){
				for(Map.Entry<Object, Object> entry : properties.entrySet()){
					out.writeStartElement(ENTRY);
					out.writeAttribute(KEY, (String)entry.getKey());
					out.writeCharacters((String)entry.getValue());
					out.writeEndElement();
				}
			}
			
			//node option subflow
			if(nodeModel.isSubflow()){
				SubflowNodeModel boxNodeModel = (SubflowNodeModel)nodeModel;
				FlowDataWriter writer = new FlowDataWriter();
				ViewLayout[] layouts = boxNodeModel.getLayouts();
				for(ViewLayout layout : layouts){
					out.writeStartElement(SUBFLOW);
					writer.write(out, new ElementWriter[] { 
							new GraphWriter(layout, null) },false, false);
					out.writeEndElement();
				}
			}
			
			out.writeEndElement();//option
			
			
			out.writeEndElement();//node
		}
		
		//edge
		for(Object ob_e : graph.getEdges()){
			FlowEdge e = (FlowEdge)ob_e;
			out.writeEmptyElement(EDGE);
			out.writeAttribute(FROM, nodeToIdMap.get(e.getSource()).toString());
			out.writeAttribute(TO, nodeToIdMap.get(e.getDest()).toString());
		}
		
		out.writeEndElement();//graph
	}

	
	Point getLayoutOffset() {
		return layout.getOffset();
	}

	/*相対位置
	
	Point2D getLocation(FlowNode node) {
		Point2D p = viewer.getGraphLayout().getLocation(node);
		p = viewer.transform(p);
		Dimension d = viewer.getSize();
		return new Point2D.Double(p.getX()/d.width, p.getY()/d.height);
	}
	*/
	
	/*絶対位置
	
	Point2D getLocation(FlowNode node) {
		Point2D p = viewer.getGraphLayout().getLocation(node);
		p = viewer.layoutTransform(p);
		return new Point2D.Float((int)p.getX(), (int)p.getY());
	}*/
	
	
	Point2D getLocation(FlowNode node) {
		return layout.getLocation(node);
	}

	
	Dimension getSize() {
		return layout.getSize();
	}

	
	String getLocationType() {
		return LOCATION_TYPE_ABSOLUTE;
	}

	
	void writeFlowSettings(XMLWriter out) throws XMLException {
		if(flowSettings == null)
			return ;
		
		out.writeStartElement(FLOWSETTING);
	
		for(Map.Entry<Object, Object> entry : flowSettings.entrySet()){
			out.writeStartElement(ENTRY);
			out.writeAttribute(KEY, (String)entry.getKey());
			out.writeCharacters((String)entry.getValue());
			out.writeEndElement();
		}
		
		out.writeEndElement();//FILESETTINGS
	}
}
