package com.ef_prime.rflow.flow;

import java.awt.geom.Point2D;
import java.util.Map;

import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEdit;

import com.ef_prime.rflow.flow.model.FlowEdge;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeProperty;
import com.ef_prime.rflow.resource.Messages;

import edu.uci.ics.jung.graph.Vertex;

class UndoableEditFactory {
	private FlowViewer viewer;

	UndoableEditFactory(FlowViewer viewer) {
		this.viewer = viewer;
	}
	
	UndoableEdit createAddNodeEdit(final FlowNode node)
	{
		return new DefaultUndoableEdits.AddNodeEdit(viewer, node);
	}
	

	UndoableEdit createDeleteEdit(
			final FlowNode[] deletedNodes,
			final FlowEdge[] deletedEdges)
	{
		return new AbstractUndoableEdit(){
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				if(deletedNodes != null)
					for(FlowNode node : deletedNodes)
						viewer.getGraph().addVertex(node);

				if(deletedEdges != null)
					for(FlowEdge edge : deletedEdges)
						viewer.getGraph().addEdge(edge);
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				if(deletedNodes != null)
					for(FlowNode node : deletedNodes)
						viewer.getGraph().removeVertex(node);
				if(deletedNodes == null || deletedNodes.length == 0)
					if(deletedEdges != null)
						for(FlowEdge edge : deletedEdges)
							viewer.getGraph().removeEdge(edge);
			}
			@Override
			public String getPresentationName() {
				if(deletedNodes != null && deletedNodes.length > 0)
					return Messages.getString("Action.remove_node"); //$NON-NLS-1$
				else
					return Messages.getString("Action.remove_edge"); //$NON-NLS-1$
			}
		};
	}
	
	
	UndoableEdit createAddEdgeEdit(final FlowEdge[] edges)
	{
		return new AbstractUndoableEdit(){
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				for(FlowEdge edge : edges)
					viewer.getGraph().removeEdge(edge);
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				for(FlowEdge edge : edges)
					viewer.getGraph().addEdge(edge);
			}
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.new_edge"); //$NON-NLS-1$
			}
		};
	}
	
	UndoableEdit createAddSublayoutEdit(final FlowNode[] nodes, final FlowEdge[] edges)
	{
		return new AbstractUndoableEdit(){
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				for(FlowNode node : nodes)
					viewer.getGraph().removeVertex(node);
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				for(FlowNode node : nodes)
					viewer.getGraph().addVertex(node);
				for(FlowEdge edge : edges)
					viewer.getGraph().addEdge(edge);
			}
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.add_subflow"); //$NON-NLS-1$
			}
		};
	}
	
	
	UndoableEdit createMoveNodeEdit(final Map<Vertex, Point2D> map){
		return new AbstractUndoableEdit(){
			
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				setLocation();
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				setLocation();
			}
			
			private void setLocation(){
				for(Map.Entry<Vertex, Point2D> entry : map.entrySet()){
					Vertex node = entry.getKey();
					Point2D p = entry.getValue();
					map.put(node, (Point2D)viewer.getGraphLayout().getLocation(node).clone());
					viewer.getGraphLayout().forceMove(node, p.getX(), p.getY());
				}
				viewer.repaint();
			}
			
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.move_node"); //$NON-NLS-1$
			}
		};
	}
	
	
	UndoableEdit createNodeModelChangeEdit(final FlowNode node, final NodeModel oldModel){
		return new AbstractUndoableEdit(){
			
			private NodeModel model = oldModel;
			
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				switchModel();
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				switchModel();
			}
			
			private void switchModel(){
				NodeModel temp = node.getModel();
				node.setModel(model);
				model = temp;
				viewer.getPickedState().clearPickedVertices();
				viewer.getPickedState().pick(node, true);
			}
			
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.edit_node"); //$NON-NLS-1$
			}
		}; 
	}
	

	UndoableEdit createNodePropertyChangeEdit(final FlowNode node, final NodeProperty oldProperty){
		return new AbstractUndoableEdit(){
			
			private NodeProperty property = oldProperty;
			
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				switchProperty();
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				switchProperty();
			}
			
			private void switchProperty(){
				NodeProperty temp = node.getProperty();
				node.setProperty(property);
				property = temp;
				viewer.getPickedState().clearPickedVertices();
				viewer.getPickedState().pick(node, true);
			}
			
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.change_property"); //$NON-NLS-1$
			}
		}; 
	}
	

	UndoableEdit createInheritStateEdit(final Map<FlowNode, FlowNode> fromToMap){
		return new AbstractUndoableEdit(){
			
			private Map<FlowNode, FlowNode> map = fromToMap;
			
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				FlowState state = viewer.getRootFlowState();
				for(Map.Entry<FlowNode, FlowNode> entry : map.entrySet()){
					FlowNode old = entry.getKey();
					FlowNode newNode = entry.getValue();
					state.passState(newNode, old);
				}
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				FlowState state = viewer.getRootFlowState();
				for(Map.Entry<FlowNode, FlowNode> entry : map.entrySet()){
					FlowNode old = entry.getKey();
					FlowNode newNode = entry.getValue();
					state.passState(old, newNode);
				}
			}
			/*private void passState(FlowState state, FlowNode from, FlowNode to) {
				if(from.isCacheData()){
					state.setCache(to, from.getCacheData());
					if(from.getCacheData().isValid()){
						to.getCacheData().setValid(true);
						from.getCacheData().setValid(false);
					}

					state.removeCacheNode(from);
				}
			}*/
		}; 
	}
	
	UndoableEdit createLockCacheStateEdit(final boolean lock){
		return new AbstractUndoableEdit(){
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				viewer.getCacheHandler().setCacheStateLock(!lock);
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				viewer.getCacheHandler().setCacheStateLock(lock);
			}
		};
	}
	
	UndoableEdit createClearCache(final FlowNode node, boolean removeMark, boolean significant){
		return new DefaultUndoableEdits.ClearCacheEdit(viewer, node, removeMark, significant);
	}
}