package com.ef_prime.rflow.flow.model;

import java.awt.Window;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Icon;

import com.ef_prime.rflow.flow.CacheData;
import com.ef_prime.rflow.flow.FlowViewer;
import com.ef_prime.rflow.group.AppearanceGroup;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeModelChangeListener;
import com.ef_prime.rflow.node.NodeModelManager;
import com.ef_prime.rflow.node.NodeProperty;
import com.ef_prime.rflow.node.NodePropertyChangeListener;
import com.ef_prime.rflow.node.base.SubflowNodeModel;

import edu.uci.ics.jung.graph.impl.DirectedSparseVertex;

public class FlowNode extends DirectedSparseVertex{
	
	//model---------------------------------------------
	private NodeModel model;
	private NodeProperty property;
	
	//other---------------------------------------------
	//followings are not copied, when copy()
	private CacheData cacheData = null;
	private Window editWindow = null;
	
	public FlowNode(NodeModel model){
		this(model, NodeProperty.createDefaultProperty());
	}
	
	public FlowNode(NodeModel model, NodeProperty property){
		super();
		this.model = model;
		this.property = property;

		if(model.isSubflow())
			((SubflowNodeModel)model).setParentNode(this);
	}
	
	/**
	 * フルコピー用のコンストラクタ
	 * @param model
	 * @param property
	 * @param cacheData
	 */
	private FlowNode(NodeModel model, NodeProperty property, CacheData cacheData){
		this(model, property);
		this.cacheData = cacheData;
	}

	public NodeModel getModel() {
		return model;
	}

	public void setModel(NodeModel model) {
		if(getModel() != null && getModel().equals(model))
			return;
		NodeModel old = this.model;
		this.model = model;
		if(model.isSubflow())
			((SubflowNodeModel)model).setParentNode(this);
		fireNodeModelChanged(this, old, model);
	}
	
	public NodeProperty getProperty(){
		return property;
	}
	
	public void setProperty(NodeProperty property){
		if(getProperty().equals(property))
			return;
		NodeProperty old = this.property;
		this.property = property;
		fireNodePropertyChanged(this, old, property);
	}
	
	public String getTitle(){
		String title = property.getTitle();
		if(title != null)
			return title;
		else
			return model.getDefaultTitle();
	}
	
	public CacheData getCacheData() {
		return cacheData;
	}
	
	public void setCacheData(CacheData cacheData) {
		this.cacheData = cacheData;
	}
	
	public boolean isCacheData(){
		return cacheData != null && cacheData.flagExists();
	}
	
	@Override
	public FlowGraph getGraph() {
		return (FlowGraph)super.getGraph();
	}
	
	public void editNode(FlowViewer viewer, boolean selectProperty, boolean isNewNode){
		if(editWindow == null){
			editWindow = NodeModelManager.getModelController(
					getModel().getClass()).createEditDialog(viewer, this, selectProperty, isNewNode);
			if(editWindow != null){
				editWindow.addWindowListener(new WindowAdapter(){
					@Override
					public void windowClosed(WindowEvent e) {
						editWindow = null;
					}
				});
				editWindow.setVisible(true);
			}
		}else{
			editWindow.toFront();
		}
	}
	
	public void clearEditWindow(){
		if(editWindow != null){
			editWindow.dispose();
			editWindow = null;
		}
	}
	
	public Icon getIcon(){
		AppearanceGroup group = getProperty().getAppearanceGroup();
		return (group == null)?getModel().getIcon():group.getIcon();
	}
	
	public Icon getSmallIcon(){
		AppearanceGroup group = getProperty().getAppearanceGroup();
		return (group == null)?getModel().getSmallIcon():group.getSmallIcon();
	}
	
	public FlowNode copy(){
		return new FlowNode(model.copy(), property.copy());
	}
	
	/**
	 * フルコピー（キャッシュ情報もコピー）
	 * @return
	 */
	public FlowNode fullCopy(){
		if(cacheData != null)
			return new FlowNode(model.copy(), property.copy(), cacheData.copy());
		else
			return new FlowNode(model.copy(), property.copy());
	}

	public static FlowNode getTopClosedSubflowNode(FlowEdge innerEdge){
		return getTopClosedSubflowNode((FlowNode)innerEdge.getSource());
	}
	public static FlowNode getTopClosedSubflowNode(FlowNode innerNode){
		FlowNode node = innerNode;
		FlowNode closed = null;
		while(!node.getGraph().isRootGraph())
		{
			node = node.getGraph().getParentNode();
			if(!node.getModel().isOpenSubflow());
				closed = node;
		}
		if(closed != null)
			return closed;
		else
			return null;
	}
	
	private static List<NodeModelChangeListener> listeners = new ArrayList<NodeModelChangeListener>();
	private static List<NodePropertyChangeListener> propertyListeners = new ArrayList<NodePropertyChangeListener>();
	
	public static void addNodeModelChangeListener(NodeModelChangeListener listener){
		listeners.add(listener);
	}
	
	public static void removeNodeModelChangeListener(NodeModelChangeListener listener){
		listeners.remove(listener);
	}
	
	private static void fireNodeModelChanged(FlowNode node, NodeModel old, NodeModel model){
		if(!model.isSubflow()){
			for(NodeModelChangeListener l : listeners){
				l.nodeModelChanged(node, old, model);
			}
			//if changed node model is in closed subflow
			//fire top closed subflow node model changed event
			/*FlowNode closed = getTopClosedSubflowNode(node);
			if(closed != null)
				fireSubflowNodeModelChanged(closed);*/
		}else{
			fireSubflowNodeModelChanged(node);
		}
		
		//fire parent model changed
		if(!node.getGraph().isRootGraph()){
			fireSubflowNodeModelChanged(node.getGraph().getParentNode());
		}
		
	}
	
	public static void fireSubflowNodeModelChanged(FlowNode node){
		for(NodeModelChangeListener l : listeners){
			l.subflowNodeModelChanged(node);
		}
		
		//fire parent model changed
		if(!node.getGraph().isRootGraph()){
			fireSubflowNodeModelChanged(node.getGraph().getParentNode());
		}
	}
	
	public static void addNodePropertyChangeListener(NodePropertyChangeListener listener){
		propertyListeners.add(listener);
	}
	
	public static void removeNodePropertyChangeListener(NodePropertyChangeListener listener){
		propertyListeners.remove(listener);
	}
	
	private static void fireNodePropertyChanged(FlowNode node, NodeProperty old, NodeProperty property){
		for(NodePropertyChangeListener l : propertyListeners){
			l.nodePropertyChanged(node, old, property);
		}
	}
}
