package com.ef_prime.rflow.node.pack;

import java.awt.Dimension;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.Properties;
import java.util.TreeSet;
import java.util.Map.Entry;

import com.ef_prime.rflow.Constants;
import com.ef_prime.rflow.exception.BuildFailedException;
import com.ef_prime.rflow.util.RTypeFilter;

public class ArgComponentManager {
	
	private Hashtable<String, AbstractArgComponentBuilder> builderMap = new Hashtable<String, AbstractArgComponentBuilder>();
	
	public ArgComponentManager(){
		builderMap.put("AssignInput", new AssignInputBuilder()); //$NON-NLS-1$
		builderMap.put("FreeInput", new FreeInputBuilder()); //$NON-NLS-1$
		builderMap.put("FileInput", new FileInputBuilder()); //$NON-NLS-1$
		builderMap.put("RadioButton", new RadioButtonBuilder()); //$NON-NLS-1$
		builderMap.put("ComboBox", new ComboBoxBuilder()); //$NON-NLS-1$
		builderMap.put("List", new ListBuilder()); //$NON-NLS-1$
		builderMap.put("DataFrameChooser", new DataFrameChooserBuilder()); //$NON-NLS-1$
		builderMap.put("SimpleFormulaEditor", new SimpleFormulaEditorBuilder()); //$NON-NLS-1$
		builderMap.put("FormulaVariableChooser", new FormulaVariableChooserBuilder()); //$NON-NLS-1$
		builderMap.put("CheckBox", new CheckBoxBuilder()); //$NON-NLS-1$
		builderMap.put("NumberInput", new NumberInputBuilder()); //$NON-NLS-1$
	}
	
	public ArgComponentBuilder getArgComponentBuilder(
			String type,
			String param, String title, Dimension size, Properties entrys)
	{
		AbstractArgComponentBuilder builder;
		builder = builderMap.get(type);
		if(builder == null)return null;
		builder = (AbstractArgComponentBuilder)builder.clone();
		builder.setData(param, title, size, entrys);
		return builder;
	}
	
	private static abstract class AbstractArgComponentBuilder implements ArgComponentBuilder ,GetBuilderInfo{
		protected String param;
		protected String title; 
		protected Dimension size; 
		protected Properties entrys;
		
		public void setData(String param, String title, Dimension size, Properties entrys) {
			this.param = param;
			this.title = title;
			this.size = size;
			this.entrys = entrys;
		}
		
		public String getDescription(){
			return ""; //$NON-NLS-1$
		}
		
		@Override
		protected Object clone(){
			try {
				return super.clone();
			} catch (CloneNotSupportedException e) {
				e.printStackTrace();throw new RuntimeException();//unreachable
			}
		}
	}
	
	private static abstract class DefaultArgComponentBuilder extends AbstractArgComponentBuilder{
		public final ArgComponent buildArgComponent() throws BuildFailedException {
			Properties entrysDef = entrys;
			Properties defaultEntrys = getDefaultProperties();
			if(defaultEntrys != null){
				entrysDef = new Properties(defaultEntrys);
				entrysDef.putAll(entrys);
			}
			ArgComponent ac = createArgComponent(entrysDef);
			if(size != null){
				ac.setMaximumSize(size);
				ac.setPreferredSize(size);
			}
			return ac;
		}
		
		protected abstract ArgComponent createArgComponent(Properties entrys) throws BuildFailedException;
		
		protected Properties getDefaultProperties(){
			return null;
		}
	}
	
	private static class AssignInputBuilder extends AbstractArgComponentBuilder{
		private static final String DEFAULT_TITLE = "Result Object Name"; //$NON-NLS-1$
		
		public OptionInfo[] getInfoList() {
			return null;
		}
		
		public ArgComponent buildArgComponent() {
			if(title == null || title.equals("")) //$NON-NLS-1$
				title = DEFAULT_TITLE;
			return DefaultArgComponents.createSymbolInput(title);
		}
	}
	
	private static class FreeInputBuilder extends DefaultArgComponentBuilder{

		
		/*----------------------------option------------------------*/
		private static final String QUOTE = "quote"; //$NON-NLS-1$
		
		/*-------------------------support---------------------------*/
		private static final String SUPPORT = "support"; //$NON-NLS-1$
		private static final String METHOD = "method"; //$NON-NLS-1$
		private static final String TARGET = "target"; //$NON-NLS-1$
		
		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{
					DefaultInfos.QUOTE,
					DefaultInfos.SUPPORT,
					DefaultInfos.METHOD,
					DefaultInfos.TARGET};
		}
		
		@Override
		protected Properties getDefaultProperties() {
			Properties p = new Properties();
			p.put(QUOTE, "true"); //$NON-NLS-1$
			p.put(SUPPORT, "false"); //$NON-NLS-1$
			p.put(METHOD, RTypeFilter.FilterMethod.Is.toString());
			
			return p;
		}
		
		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			boolean support = Boolean.parseBoolean(entrys.getProperty(SUPPORT));

			if(!support){
				boolean quote = Boolean.parseBoolean(entrys.getProperty(QUOTE));
			
				return DefaultArgComponents.createFreeInput(param, title, quote);
			}else{
				RTypeFilter filter = createFilter(entrys.getProperty(METHOD), entrys.getProperty(TARGET));
				
				return DefaultArgComponents.createSupportedFreeInput(param, title, filter);
			}
		}
	}
	
	private static class FileInputBuilder extends DefaultArgComponentBuilder{

		private static final String QUOTE = "quote"; //$NON-NLS-1$
		
		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{DefaultInfos.QUOTE};
		}

		@Override
		protected Properties getDefaultProperties() {
			Properties p = new Properties();
			p.put(QUOTE, "true"); //$NON-NLS-1$
			return p;
		}
		
		protected ArgComponent createArgComponent(Properties entrys) {
			boolean quote = Boolean.parseBoolean(entrys.getProperty(QUOTE));
			
			return DefaultArgComponents.createFileInput(param, title, quote);
		}
	}
	
	private static class RadioButtonBuilder extends DefaultArgComponentBuilder{

		private static final String QUOTE = "quote"; //$NON-NLS-1$
		private static final String LABEL = "label"; //$NON-NLS-1$
		private static final String VALUE = "value"; //$NON-NLS-1$
		private static final String AXIS = "axis"; //$NON-NLS-1$
		
		private static final String FREEINPUT = "[freeinput]"; //$NON-NLS-1$
		
		private static final String X_AXIS = "x"; //$NON-NLS-1$
		//private static final String Y_AXIS = "y";

		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{DefaultInfos.QUOTE, DefaultInfos.LABEL,
					new OptionInfo(DefaultInfos.VALUE.getKey(), DefaultInfos.VALUE.getValue(), 
							DefaultInfos.VALUE.getDescription()+" :freeinput if value is [freeinput]"), //$NON-NLS-1$
					new OptionInfo("axis", new String[]{"x", "y"}, "axis of button layout")}; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
		}
		
		@Override
		protected Properties getDefaultProperties() {
			Properties p = new Properties();
			p.put(QUOTE, "true"); //$NON-NLS-1$
			p.put(AXIS, X_AXIS);
			return p;
		}
		
		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			boolean quote = Boolean.parseBoolean(entrys.getProperty(QUOTE));
			String label = entrys.getProperty(LABEL);
			String value = entrys.getProperty(VALUE);
			if(label == null || value == null)
				throw new BuildFailedException();
			
			String[] values = value.split(Constants.DELIM);
			for(int i = 0; i < values.length; i++){
				if(values[i].equals(FREEINPUT))
					values[i] = null;
				else{
					if(quote)
						values[i] = Constants.quote(values[i]);
				}
			}
			
			boolean x_axis = entrys.getProperty(AXIS).equals(X_AXIS);
			
			return DefaultArgComponents.createRadioButton(param, title, 
					label.split(Constants.DELIM), values, quote, x_axis);
		}
	}
	

	private static class ComboBoxBuilder extends DefaultArgComponentBuilder{

		private static final String EDITABLE = "editable"; //$NON-NLS-1$
		/*-------------------------non object---------------------------*/
		private static final String LABEL = "label"; //$NON-NLS-1$
		private static final String VALUE = "value"; //$NON-NLS-1$

		/*-------------------------object---------------------------*/
		private static final String OBJECT = "object"; //$NON-NLS-1$
		private static final String METHOD = "method"; //$NON-NLS-1$
		private static final String TARGET = "target"; //$NON-NLS-1$
		
		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{new OptionInfo(EDITABLE, DefaultInfos.BOOLEAN_VALUES, "editable combobox if true"), //$NON-NLS-1$
					DefaultInfos.LABEL, DefaultInfos.VALUE, DefaultInfos.OBJECT, DefaultInfos.METHOD, DefaultInfos.TARGET};
		}
		
		@Override
		protected Properties getDefaultProperties() {
			Properties p = new Properties();
			p.put(OBJECT, "false"); //$NON-NLS-1$
			p.put(EDITABLE, "false"); //$NON-NLS-1$
			p.put(METHOD, RTypeFilter.FilterMethod.Is.toString());
			
			return p;
		}
		
		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			boolean object = Boolean.parseBoolean(entrys.getProperty(OBJECT));
			boolean editable = Boolean.parseBoolean(entrys.getProperty(EDITABLE));
			if(!object){
				String label = entrys.getProperty(LABEL);
				String value = entrys.getProperty(VALUE);
				if(label == null || value == null)
					throw new BuildFailedException();
				
				return DefaultArgComponents.createComboBox(param, title, 
						label.split(Constants.DELIM), value.split(Constants.DELIM),editable);
			}else{
				RTypeFilter filter = createFilter(entrys.getProperty(METHOD), entrys.getProperty(TARGET));
				
				return DefaultArgComponents.createObjectComboBox(param, title, filter, editable);
			}
		}
	}
	

	private static class ListBuilder extends DefaultArgComponentBuilder{

		private static final String SELECTION = "selection"; //$NON-NLS-1$
		private static final String SELECTION_SINGLE = "single"; //$NON-NLS-1$
		//private static final String SELECTION_MULTI = "multi";
		private static final String LABEL = "label"; //$NON-NLS-1$
		private static final String VALUE = "value"; //$NON-NLS-1$
		/*-------------------------object---------------------------*/
		private static final String OBJECT = "object"; //$NON-NLS-1$
		private static final String METHOD = "method"; //$NON-NLS-1$
		private static final String TARGET = "target"; //$NON-NLS-1$
		
		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{DefaultInfos.SELECTION, DefaultInfos.LABEL, DefaultInfos.VALUE,
					DefaultInfos.OBJECT, DefaultInfos.METHOD, DefaultInfos.TARGET};
		}
		
		@Override
		protected Properties getDefaultProperties() {
			Properties p = new Properties();
			p.put(SELECTION, SELECTION_SINGLE);
			p.put(OBJECT, "false"); //$NON-NLS-1$
			p.put(METHOD, RTypeFilter.FilterMethod.Is.toString());
			return p;
		}
		
		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			boolean singleSelection;
			String selection = entrys.getProperty(SELECTION);
			singleSelection = selection.equals(SELECTION_SINGLE);
			
			boolean object = Boolean.parseBoolean(entrys.getProperty(OBJECT));
			if(!object){
				String label = entrys.getProperty(LABEL);
				String value = entrys.getProperty(VALUE);
				if(label == null || value == null)
					throw new BuildFailedException();

				return DefaultArgComponents.createList(param, title, 
						label.split(Constants.DELIM), value.split(Constants.DELIM), singleSelection);
			}else{

				RTypeFilter filter = createFilter(entrys.getProperty(METHOD), entrys.getProperty(TARGET));
				
				return DefaultArgComponents.createObjectList(param, title, filter, singleSelection);
			}
		}
	}


	private static class DataFrameChooserBuilder extends DefaultArgComponentBuilder{

		public OptionInfo[] getInfoList() {
			return null;
		}
		
		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			return DefaultArgComponents.createDataFrameChooser(param);
		}
	}
	
	private static class SimpleFormulaEditorBuilder extends DefaultArgComponentBuilder{
		
		private static final String ARG_FORMULA = "formula"; //$NON-NLS-1$
		private static final String ARG_DATA = "data"; //$NON-NLS-1$
		private static final String TITLE1 = "title1"; //$NON-NLS-1$
		private static final String TITLE2 = "title2"; //$NON-NLS-1$
		
		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{DefaultInfos.FORMULA, DefaultInfos.DATA,
					new OptionInfo(TITLE1, "()", "title of response variable"), //$NON-NLS-1$ //$NON-NLS-2$
					new OptionInfo(TITLE2, "()", "title of explanatory variables")}; //$NON-NLS-1$ //$NON-NLS-2$
		}
		
		@Override
		protected Properties getDefaultProperties() {
			Properties p = new Properties();
			p.put(ARG_FORMULA, "formula"); //$NON-NLS-1$
			p.put(ARG_DATA, "data"); //$NON-NLS-1$
			p.put(TITLE1, "response"); //$NON-NLS-1$
			p.put(TITLE2, "explanatory variables"); //$NON-NLS-1$
			return p;
		}

		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			return DefaultArgComponents.createSimpleFormulaEditor(
					entrys.getProperty(ARG_FORMULA), entrys.getProperty(ARG_DATA),
					title, entrys.getProperty(TITLE1), entrys.getProperty(TITLE2));
		}
	}
	

	private static class FormulaVariableChooserBuilder extends DefaultArgComponentBuilder{
		
		private static final String ARG_FORMULA = "formula"; //$NON-NLS-1$
		private static final String ARG_DATA = "data"; //$NON-NLS-1$
		/*-------------------------selection--------------------------*/
		private static final String SELECTION = "selection"; //$NON-NLS-1$
		//private static final String SELECTION_SINGLE = "single";
		private static final String SELECTION_MULTI = "multi"; //$NON-NLS-1$
		
		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{DefaultInfos.FORMULA, DefaultInfos.DATA, DefaultInfos.SELECTION};
		}
		
		@Override
		protected Properties getDefaultProperties() {
			Properties p = new Properties();
			p.put(ARG_FORMULA, "formula"); //$NON-NLS-1$
			p.put(ARG_DATA, "data"); //$NON-NLS-1$
			p.put(SELECTION, SELECTION_MULTI);
			return p;
		}

		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			boolean singleSelection = !entrys.getProperty(SELECTION).equals(SELECTION_MULTI);
			
			return DefaultArgComponents.createFormulaVariableChooser(
					entrys.getProperty(ARG_FORMULA), entrys.getProperty(ARG_DATA),
					title, singleSelection);
		}
	}

	private static class CheckBoxBuilder extends DefaultArgComponentBuilder{
		
		private static final String LABEL = "label"; //$NON-NLS-1$

		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{new OptionInfo(LABEL, "()", "label of checkbox")}; //$NON-NLS-1$ //$NON-NLS-2$
		}
		
		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			String label = entrys.getProperty(LABEL);
			if(label == null)
				label = title;
			return DefaultArgComponents.createCheckBox(param, label);
		}
		
		@Override
		public String getDescription() {
			return "label=title if label is emtpy."; //$NON-NLS-1$
		}
	}

	private static class NumberInputBuilder extends DefaultArgComponentBuilder{

		private static final String LABEL = "label"; //$NON-NLS-1$
		private static final String INT = "int"; //$NON-NLS-1$
		private static final String MIN = "min"; //$NON-NLS-1$
		private static final String MAX = "max"; //$NON-NLS-1$
		private static final String STEP = "step"; //$NON-NLS-1$

		public OptionInfo[] getInfoList() {
			return new OptionInfo[]{new OptionInfo(LABEL, "()", "label of numberinput"), //$NON-NLS-1$ //$NON-NLS-2$
					new OptionInfo(INT, DefaultInfos.BOOLEAN_VALUES, "true if int value"), //$NON-NLS-1$
					new OptionInfo(MIN, "(number)", "min of value"), //$NON-NLS-1$ //$NON-NLS-2$
					new OptionInfo(MAX, "(number)", "max of value"), //$NON-NLS-1$ //$NON-NLS-2$
					new OptionInfo(STEP, "(number)", "one step of value when up/down button is pressed")}; //$NON-NLS-1$ //$NON-NLS-2$
		}
		
		@Override
		protected Properties getDefaultProperties() {
			Properties p = new Properties();
			p.put(INT, "true"); //$NON-NLS-1$
			p.put(MIN, "0"); //$NON-NLS-1$
			p.put(MAX, String.valueOf(Integer.MAX_VALUE));
			p.put(STEP, "1"); //$NON-NLS-1$
			return p;
		}
		
		protected ArgComponent createArgComponent(Properties entrys) throws BuildFailedException {
			boolean isInt = Boolean.parseBoolean(entrys.getProperty(INT));
			String label = entrys.getProperty(LABEL);
			if(label == null)
				label = title;
			Number min;
			Number max;
			Number step;
			try{
				if(isInt){
					min = Integer.parseInt(entrys.getProperty(MIN));
					max = Integer.parseInt(entrys.getProperty(MAX));
					step = Integer.parseInt(entrys.getProperty(STEP));
				}else{
					min = Double.parseDouble(entrys.getProperty(MIN));
					max = Double.parseDouble(entrys.getProperty(MAX));
					step = Double.parseDouble(entrys.getProperty(STEP));
				}
			}catch(NumberFormatException e){
				throw new BuildFailedException();
			}
			
			return DefaultArgComponents.createNumberSpinner(param, label, min, max, step, isInt);
		}
		

		@Override
		public String getDescription() {
			return "label=title if label is emtpy."; //$NON-NLS-1$
		}
	}

	private static RTypeFilter createFilter(String method, String target) throws BuildFailedException {
		if(target == null || target.equals("")){ //$NON-NLS-1$
			return RTypeFilter.NullFilter;
		}
		RTypeFilter.FilterMethod filterMethod = RTypeFilter.FilterMethod.toFilterMethod(method);
		if(filterMethod == null || target == null)throw new BuildFailedException();
		String[] targets = target.split(Constants.DELIM);

		RTypeFilter filter = new RTypeFilter();
		for(String t : targets){
			filter.addFilter(new RTypeFilter.FilterEntry(filterMethod, t));
		}
		return filter;
	}
	
	
	
	
	
	
	//-----------------------------------------------------------------------------
	//Builder Infomation
	//-----------------------------------------------------------------------------
	
	private static interface GetBuilderInfo{
		OptionInfo[] getInfoList();
	}
	
	private static class OptionInfo{
		private String key;
		private String value;
		private String description;

		public OptionInfo(String key, String[] values, String description) {
			this(key, Constants.combine(values), description);
		}
		
		public OptionInfo(String key, String value, String description) {
			super();
			this.key = key;
			this.value = value;
			this.description = description;
		}
		
		public String getKey() {
			return key;
		}
		
		public String getValue() {
			return value;
		}
		
		public String getDescription() {
			return description;
		}
	}
	
	
	private static class DefaultInfos{
		private static final String[] BOOLEAN_VALUES = new String[]{"true", "false"}; //$NON-NLS-1$ //$NON-NLS-2$
		
		private static final OptionInfo QUOTE = new OptionInfo("quote", BOOLEAN_VALUES, "value will be unquoted if true"); //$NON-NLS-1$ //$NON-NLS-2$
		private static final OptionInfo SUPPORT = new OptionInfo("support", BOOLEAN_VALUES, "enable word support if true"); //$NON-NLS-1$ //$NON-NLS-2$
		private static final OptionInfo OBJECT = new OptionInfo("object", BOOLEAN_VALUES, "set as an object list if true"); //$NON-NLS-1$ //$NON-NLS-2$
		private static final OptionInfo METHOD = new OptionInfo("method", RTypeFilter.FilterMethod.nameValues(), "filter method of 'support' or 'object' list"); //$NON-NLS-1$ //$NON-NLS-2$
		private static final OptionInfo TARGET = new OptionInfo("target", "LIST:()", "filter target of 'method'; multiple targets will filter 'or' of targets"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		private static final OptionInfo VALUE = new OptionInfo("value", "LIST:()", "selectable values"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		private static final OptionInfo LABEL = new OptionInfo("label", "LIST:()", "label of 'value'"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		private static final OptionInfo SELECTION = new OptionInfo("selection", new String[]{"single", "multi"}, "selection type"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
		private static final OptionInfo FORMULA = new OptionInfo("formula", "()", "argument name for formula"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		private static final OptionInfo DATA = new OptionInfo("data", "()", "argument name for data"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}
	
	public static void main(String[] args) {
		ArgComponentManager manager = new ArgComponentManager();
		Comparator<Entry<String, AbstractArgComponentBuilder>> comp = new Comparator<Entry<String,AbstractArgComponentBuilder>>(){
			public int compare(Entry<String, AbstractArgComponentBuilder> o1, Entry<String, AbstractArgComponentBuilder> o2) {
				return o1.getKey().compareTo(o2.getKey());
			}
		};
		TreeSet<Entry<String, AbstractArgComponentBuilder>> treeSet = new TreeSet<Entry<String,AbstractArgComponentBuilder>>(comp);
		treeSet.addAll(manager.builderMap.entrySet());
		
		for(Entry<String, AbstractArgComponentBuilder> entry : treeSet){
			//System.out.println("type: " + entry.getKey());
			System.out.println(":"+entry.getKey()+": "); //$NON-NLS-1$ //$NON-NLS-2$
			AbstractArgComponentBuilder builder = entry.getValue();
			Properties p = null;
			if (builder instanceof DefaultArgComponentBuilder) {
				DefaultArgComponentBuilder defBuilder = (DefaultArgComponentBuilder) builder;
				p = defBuilder.getDefaultProperties();
			}
			
			System.out.println(builder.getDescription());
			System.out.println();
			
			System.out.println(",key,value,default,description"); //$NON-NLS-1$
			OptionInfo[] infos = builder.getInfoList();
			if(infos != null)
				for(OptionInfo info : infos){
					String def = (p == null)?null:p.getProperty(info.getKey());
					System.out.print(","+Constants.quote(info.getKey())); //$NON-NLS-1$
					System.out.print(","); //$NON-NLS-1$
					System.out.print(Constants.quote(info.getValue()));
					System.out.print(","); //$NON-NLS-1$
					if(def != null){
						System.out.print(def);
					}else{
						
					}
					System.out.print(","); //$NON-NLS-1$
					System.out.println(Constants.quote(info.getDescription()));
				}
			System.out.println();
		}
	}
}
