package com.ef_prime.rflow.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

public class TypeDefProperties extends Properties{
	
	public static enum Type{
		STRING,
		BOOLEAN,
		INTEGER,
		FLOAT,
		DOUBLE,
		LIST_STRING(Type.STRING),
		LIST_BOOLEAN(Type.BOOLEAN),
		LIST_INTEGER(Type.INTEGER),
		LIST_FLOAT(Type.FLOAT),
		LIST_DOUBLE(Type.DOUBLE);
		
		private Type generics;
		
		private Type(){
		}
		
		private Type(Type generics){
			this.generics = generics;
		}
	}
	
	public static interface Key{
		Type getType();
		Object getDefaultValue();
		String getKeyString();
	}
	
	
	public synchronized Object setProperty(Key key, Object value) {
		//if list, check for null
		if(key.getType().generics != null){
			if(((Collection)value).contains(null))
				throw new NullPointerException();
		}
		
		Object o = setProperty(key.getKeyString(), toString(key.getType(), value));
		
		try{
			return (o == null)?null:valueOf(key.getType(), (String)o);
		}catch(Exception e){
			//if Type was incorrect
			return null;
		}
	}
	

	public synchronized Object getProperty(Key key) {
		String value = super.getProperty(key.getKeyString());
		if(value == null)
			return key.getDefaultValue();
		try{
			return valueOf(key.getType(), value);
		}catch(Exception e){
			//if Type was incorrect
			//set and return default value
			setProperty(key, key.getDefaultValue());
			return key.getDefaultValue();
		}
	}
	
	public synchronized Object remove(Key key) {
		return super.remove(key.getKeyString());
	}
	
	public synchronized boolean containsKey(Key key) {
		return super.containsKey(key.getKeyString());
	}
	
	private static String toString(Type type, Object value){
		if(value == null)return null;
		try{
			switch(type){
			case STRING:return value.toString();
			case BOOLEAN:return Boolean.toString((Boolean)value);
			case INTEGER:return Integer.toString((Integer)value);
			case FLOAT:return Float.toString((Float)value);
			case DOUBLE:return Double.toString((Double)value);
			case LIST_STRING:
			case LIST_BOOLEAN:
			case LIST_INTEGER:
			case LIST_FLOAT:
			case LIST_DOUBLE:
				return listToString(type, value);
			}
		}catch(Exception e){
			throw new IllegalArgumentException("the value Type don't match the Type of key"); //$NON-NLS-1$
		}
		throw new RuntimeException("Undefined type:"+type.toString()); //$NON-NLS-1$
	}
	
	private static Object valueOf(Type type, String value){
		switch(type){
		case STRING:return value;
		case BOOLEAN:return Boolean.valueOf(value);
		case INTEGER:return Integer.valueOf(value);
		case FLOAT:return Float.valueOf(value);
		case DOUBLE:return Double.valueOf(value);
		case LIST_STRING:
		case LIST_BOOLEAN:
		case LIST_INTEGER:
		case LIST_FLOAT:
		case LIST_DOUBLE:
			return listValueOf(type, value);
		}
		throw new RuntimeException("Undefined type:"+type.toString()); //$NON-NLS-1$
	}
	
	
	private static final String COMMA_MARK = "[:comma;]"; //$NON-NLS-1$
	
	private static String listToString(Type type, Object value){
		String ret = ""; //$NON-NLS-1$
		for(Object o : (List)value){
			String s = toString(type.generics, o);
			s = s.replaceAll(",", COMMA_MARK); //$NON-NLS-1$
			ret += s+","; //$NON-NLS-1$
		}
		
		return ret.substring(0, ret.length()-1);
	}
	
	private static Object listValueOf(Type type, String value){
		String[] elements = value.split(","); //$NON-NLS-1$
		ArrayList<Object> list = new ArrayList<Object>();
		for(String element : elements){
			String elem = element.replaceAll(COMMA_MARK, ","); //$NON-NLS-1$
			list.add(valueOf(type.generics, elem));
		}
		return list;
	}
	
	/**
	 * Propertiesにキーのデフォルト値を登録します。
	 * @param properties the Properties to set defaults
	 * @param keys all Key of this Properties
	 * @return properties  
	 */
	public static TypeDefProperties setDefaultProperties(TypeDefProperties properties, Key[] keys){
		for(Key key : keys)
			properties.setProperty(key, key.getDefaultValue());
		return properties;
	}

	/**
	 * 保存されていた<code>Properties</code>と現在の<code>Properties</code>クラスのバージョンの違いに
	 * 対応するための更新メソッドです。
	 * <p>・現在使用されていないキーの消去
	 * <p>・新しく追加されたキーをデフォルト値で登録
	 * 
	 * @param keys all Key of this Properties
	 */
	public void update(Key[] keys){
		HashSet<Object> keySet = new HashSet<Object>(keySet());
		for(Key key : keys){
			if(!containsKey(key))
				setProperty(key, key.getDefaultValue());			
			keySet.remove(key.toString());
		}
		Iterator<Object> i = keySet.iterator();
		while(i.hasNext()){
			remove((String)i.next());
		}
	}
	
	/**
	 * <code>properties</code>の中身をすべて登録します。
	 * @param properties
	 */
	public void setPropertyAll(TypeDefProperties properties){
		Enumeration<Object> keys = properties.keys();
		while(keys.hasMoreElements()){
			String key = (String)keys.nextElement();
			setProperty(key, properties.getProperty(key));
		}
	}
}
